def celsius_to_kelvin(temperature):
    return round(temperature + 273.15, 2)


def eighths_to_percent(eighths):
    return int(eighths / 8 * 100)


def hpa_to_pa(pressure):
    return int(pressure * 100)


def kelvin_to_celsius(temperature):
    return round(temperature - 273.15, 2)


def kmh_to_ms(speed):
    return round(speed / 3.6, 1)


def minutes_to_seconds(duration):
    return int(duration * 60)


def ms_to_kmh(speed):
    return round(speed * 3.6, 1)


def pa_to_hpa(pressure):
    return pressure / 100


def seconds_to_minutes(duration):
    return duration / 60


# XXX: Subsequent codes with the same mapping are left out. I.e. all codes from
#      0 through 16 map to 'dry', etc.
SYNOP_CURRENT_CONDITION_MAP = {
    0: 'dry',
    17: 'thunderstorm',
    18: 'dry',
    40: 'fog',
    50: 'rain',
    56: 'sleet',
    58: 'rain',
    66: 'sleet',
    70: 'snow',
    80: 'rain',
    83: 'sleet',
    85: 'snow',
    87: 'sleet',
    89: 'hail',
    91: 'rain',
    93: 'snow',
    95: 'thunderstorm',
    100: 'dry',
    130: 'fog',
    140: 'rain',
    145: 'snow',
    147: 'sleet',
    150: 'rain',
    154: 'sleet',
    157: 'rain',
    164: 'sleet',
    170: 'snow',
    174: 'hail',
    180: 'rain',
    185: 'sleet',
    189: 'hail',
    190: 'thunderstorm',
    193: 'hail',
    194: 'thunderstorm',
    196: 'hail',
    197: None,
    199: 'thunderstorm',
    200: None,
}


SYNOP_PAST_CONDITION_MAP = {
    0: 'dry',
    4: 'fog',
    5: 'rain',
    7: 'snow',
    8: 'rain',
    9: 'thunderstorm',
    10: 'dry',
    11: 'fog',
    14: 'rain',
    17: 'snow',
    18: 'rain',
    19: 'thunderstorm',
    20: None,
}


SYNOP_FORM_OF_PRECIPITATION_CONDITION_MAP = {
    0: 'dry',
    6: 'rain',
    7: 'snow',
    8: 'sleet',
    9: 'rain',
    10: 'dry',
    11: 'rain',
    12: 'snow',
    13: 'sleet',
    14: 'rain',
}


CURRENT_OBSERVATIONS_CONDITION_MAP = {
    1: 'dry',
    5: 'fog',
    7: 'rain',
    10: 'sleet',
    14: 'snow',
    18: 'rain',
    20: 'sleet',
    22: 'snow',
    26: 'thunderstorm',
    29: 'hail',
    31: 'dry',
    32: None,
}


def _find(mapping, code):
    if code is None:
        return
    value = None
    for k, v in mapping.items():
        if k > code:
            return value
        value = v


def synop_current_weather_code_to_condition(code):
    return _find(SYNOP_CURRENT_CONDITION_MAP, code)


def synop_past_weather_code_to_condition(code):
    return _find(SYNOP_PAST_CONDITION_MAP, code)


def synop_form_of_precipitation_code_to_condition(code):
    return _find(SYNOP_FORM_OF_PRECIPITATION_CONDITION_MAP, code)


def current_observations_weather_code_to_condition(code):
    return _find(CURRENT_OBSERVATIONS_CONDITION_MAP, code)


CONVERTERS = {
    'dwd': {
        'dew_point': kelvin_to_celsius,
        'pressure_msl': pa_to_hpa,
        'sunshine': seconds_to_minutes,
        'sunshine_10': seconds_to_minutes,
        'sunshine_30': seconds_to_minutes,
        'sunshine_60': seconds_to_minutes,
        'temperature': kelvin_to_celsius,
        'wind_speed': ms_to_kmh,
        'wind_speed_10': ms_to_kmh,
        'wind_speed_30': ms_to_kmh,
        'wind_speed_60': ms_to_kmh,
        'wind_gust_speed': ms_to_kmh,
        'wind_gust_speed_10': ms_to_kmh,
        'wind_gust_speed_30': ms_to_kmh,
        'wind_gust_speed_60': ms_to_kmh,
    }
}


def convert_record(record, units):
    for field, converter in CONVERTERS[units].items():
        if record.get(field) is not None:
            record[field] = converter(record[field])



import time

from huey import crontab, PriorityRedisHuey
from huey.api import TaskLock as TaskLock_
from huey.exceptions import TaskLockedException

from brightsky import tasks
from brightsky.settings import settings


class ExpiringLocksHuey(PriorityRedisHuey):

    def lock_task(self, lock_name):
        return TaskLock(self, lock_name)

    def expire_locks(self, timeout):
        expired = set()
        threshold = time.time() - timeout
        for key in list(self._locks):
            value = self.get(key, peek=True)
            if value and float(value) < threshold:
                self.delete(key)
                expired.add(key)
        return expired

    def is_locked(self, lock_name):
        return TaskLock(self, lock_name).is_locked()


class TaskLock(TaskLock_):

    def __enter__(self):
        if not self._huey.put_if_empty(self._key, str(time.time())):
            raise TaskLockedException('unable to set lock: %s' % self._name)

    def is_locked(self):
        return self._huey.storage.has_data_for_key(self._key)


huey = ExpiringLocksHuey(
    'brightsky',
    results=False,
    url=settings.REDIS_URL,
)


@huey.task()
def process(url):
    with huey.lock_task(url):
        tasks.parse(url=url, export=True)


@huey.periodic_task(
    crontab(minute=settings.POLLING_CRONTAB_MINUTE), priority=50)
def poll():
    tasks.poll(enqueue=True)


@huey.periodic_task(crontab(minute='23'), priority=0)
def clean():
    tasks.clean()



import os

from brightsky.cli import cli
from brightsky.utils import configure_logging, load_dotenv


if __name__ == '__main__':
    load_dotenv()
    configure_logging()
    if os.getenv('SENTRY_DSN'):
        import sentry_sdk
        sentry_sdk.init(os.getenv('SENTRY_DSN'))
    cli(prog_name='python -m brightsky')



import glob
import logging
import os
import re
from contextlib import contextmanager, suppress
from multiprocessing import cpu_count

import psycopg2
from psycopg2.extras import DictCursor
from psycopg2.pool import ThreadedConnectionPool

from brightsky.settings import settings


logger = logging.getLogger(__name__)


@contextmanager
def get_connection():
    if not hasattr(get_connection, '_pool'):
        if 'gunicorn' in os.getenv('SERVER_SOFTWARE', ''):
            # gunicorn sync workers are single-threaded
            minconn = 1
        else:
            minconn = 2*cpu_count()+1
        get_connection._pool = ThreadedConnectionPool(
            minconn, minconn, settings.DATABASE_URL, cursor_factory=DictCursor)
    pool = get_connection._pool
    conn = pool.getconn()
    try:
        with conn:
            yield conn
    except psycopg2.InterfaceError:
        logger.warning('Discarding dead connection pool')
        pool.closeall()
        del get_connection._pool
        raise
    finally:
        if not pool.closed:
            pool.putconn(conn)


def fetch(*args, **kwargs):
    for retry in range(5):
        with suppress(psycopg2.InterfaceError):
            with get_connection() as conn:
                with conn.cursor() as cur:
                    cur.execute(*args, **kwargs)
                    return cur.fetchall()


def migrate():
    logger.info("Migrating database")
    with get_connection() as conn:
        with conn.cursor() as cur:
            try:
                cur.execute('SELECT MAX(id) FROM migrations;')
            except psycopg2.errors.UndefinedTable:
                conn.rollback()
                latest_migration = 0
            else:
                latest_migration = cur.fetchone()[0]

            migration_paths = [
                f for f in sorted(glob.glob('migrations/*.sql'))
                if (m := re.match(r'(\d+)_', os.path.basename(f)))
                and int(m.group(1)) > latest_migration
            ]

            for path in migration_paths:
                logger.info("Applying %s", path)
                match = re.match(r'(\d+)_?(.*)\.sql', os.path.basename(path))
                migration_id = int(match.group(1))
                migration_name = match.group(2)
                with open(path) as f:
                    cur.execute(f.read())
                cur.execute(
                    'INSERT INTO migrations (id, name) VALUES (%s, %s);',
                    (migration_id, migration_name))
                conn.commit()
        if migration_paths:
            logger.info("Applied %d migrations", len(migration_paths))
        else:
            logger.info("No new migrations")



import datetime

from brightsky.db import fetch


def _make_dicts(rows):
    return [dict(row) for row in rows]


def weather(
        date, last_date=None, lat=None, lon=None, dwd_station_id=None,
        wmo_station_id=None, source_id=None, max_dist=50000):
    if not last_date:
        last_date = date + datetime.timedelta(days=1)
    sources_rows = sources(
        lat=lat, lon=lon, dwd_station_id=dwd_station_id, source_id=source_id,
        wmo_station_id=wmo_station_id, max_dist=max_dist,
        observation_types=['historical', 'recent', 'current', 'forecast'],
        date=date, last_date=last_date)['sources']
    primary_source_ids = {}
    for row in sources_rows:
        primary_source_ids.setdefault(row['observation_type'], row['id'])
    primary_source_ids = list(primary_source_ids.values())
    weather_rows = _weather(date, last_date, primary_source_ids)
    source_ids = [row['id'] for row in sources_rows]
    if len(weather_rows) < int((last_date - date).total_seconds()) // 3600:
        weather_rows = _weather(date, last_date, source_ids)
    _fill_missing_fields(weather_rows, date, last_date, source_ids)
    used_source_ids = {row['source_id'] for row in weather_rows}
    used_source_ids.update(
        source_id
        for row in weather_rows
        for source_id in row.get('fallback_source_ids', {}).values())
    return {
        'weather': weather_rows,
        'sources': [s for s in sources_rows if s['id'] in used_source_ids],
    }


def _weather(date, last_date, source_id, not_null=None):
    params = {
        'date': date,
        'last_date': last_date,
        'source_id': source_id,
    }
    where = "timestamp BETWEEN %(date)s AND %(last_date)s"
    order_by = "timestamp"
    if isinstance(source_id, list):
        where += " AND source_id IN %(source_id_tuple)s"
        order_by += ", array_position(%(source_id)s, source_id)"
        params['source_id_tuple'] = tuple(source_id)
    else:
        where += " AND source_id = %(source_id)s"
    if not_null:
        where += ''.join(f" AND {element} IS NOT NULL" for element in not_null)
    sql = f"""
        SELECT DISTINCT ON (timestamp) *
        FROM weather
        WHERE {where}
        ORDER BY {order_by}
    """
    return _make_dicts(fetch(sql, params))


# Not available in MOSMIX
IGNORED_MISSING_FIELDS = {'wind_gust_direction', 'relative_humidity'}


def _fill_missing_fields(weather_rows, date, last_date, source_ids):
    incomplete_rows = []
    missing_fields = set()
    for row in weather_rows:
        missing_row_fields = set(
            k for k, v in row.items() if v is None
        ).difference(IGNORED_MISSING_FIELDS)
        if missing_row_fields:
            incomplete_rows.append((row, missing_row_fields))
            missing_fields.update(missing_row_fields)
    if incomplete_rows:
        min_date = incomplete_rows[0][0]['timestamp']
        max_date = incomplete_rows[-1][0]['timestamp']
        # NOTE: If there are multiple missing fields we may be missing out on
        #       a "better" fallback if there are preferred sources that have
        #       one (but not all) of the missing fields. However, this lets us
        #       get by with using the basic weather query, and with performing
        #       it only one extra time.
        fallback_rows = {
            row['timestamp']: row
            for row in _weather(
                min_date, max_date, source_ids, not_null=missing_fields)
        }
        for row, fields in incomplete_rows:
            fallback_row = fallback_rows.get(row['timestamp'])
            if fallback_row:
                row['fallback_source_ids'] = {}
                for f in fields:
                    row[f] = fallback_row[f]
                    row['fallback_source_ids'][f] = fallback_row['source_id']


def current_weather(
        lat=None, lon=None, dwd_station_id=None, wmo_station_id=None,
        source_id=None, max_dist=50000, fallback=True):
    sources_rows = sources(
        lat=lat, lon=lon, dwd_station_id=dwd_station_id,
        wmo_station_id=wmo_station_id, observation_types=['synop'],
        max_dist=max_dist
    )['sources']
    source_ids = [row['id'] for row in sources_rows]
    weather = _current_weather(source_ids)
    if not weather:
        raise LookupError(
            "Could not find current weather for your location criteria")
    used_source_ids = [weather['source_id']]
    if fallback:
        missing_fields = [k for k, v in weather.items() if v is None]
        fallback_weather = _current_weather(
            source_ids, not_null=missing_fields)
        if fallback_weather:
            weather.update({k: fallback_weather[k] for k in missing_fields})
            weather['fallback_source_ids'] = {
                field: fallback_weather['source_id']
                for field in missing_fields}
            used_source_ids.append(fallback_weather['source_id'])
    return {
        'weather': weather,
        'sources': [s for s in sources_rows if s['id'] in used_source_ids],
    }


def _current_weather(source_ids, not_null=None):
    params = {
        'source_ids': source_ids,
        'source_ids_tuple': tuple(source_ids),
    }
    where = "source_id IN %(source_ids_tuple)s"
    if not_null:
        where += ''.join(f" AND {element} IS NOT NULL" for element in not_null)
    sql = f"""
        SELECT *
        FROM current_weather
        WHERE {where}
        ORDER BY array_position(%(source_ids)s, source_id)
        LIMIT 1
    """
    rows = _make_dicts(fetch(sql, params))
    if not rows:
        return {}
    return rows[0]


def synop(
        date, last_date=None, dwd_station_id=None, wmo_station_id=None,
        source_id=None):
    if not last_date:
        last_date = date + datetime.timedelta(days=1)
    sources_rows = sources(
        dwd_station_id=dwd_station_id, wmo_station_id=wmo_station_id,
        source_id=source_id, observation_types=['synop'])['sources']
    source_ids = [row['id'] for row in sources_rows]
    sql = """
        SELECT *
        FROM synop
        WHERE
            timestamp BETWEEN %(date)s AND %(last_date)s AND
            source_id IN %(source_ids_tuple)s
        ORDER BY timestamp
        """
    params = {
        'date': date,
        'last_date': last_date,
        'source_ids_tuple': tuple(source_ids),
    }
    return {
        'weather': _make_dicts(fetch(sql, params)),
        'sources': _make_dicts(sources_rows),
    }


def sources(
        lat=None, lon=None, dwd_station_id=None, wmo_station_id=None,
        source_id=None, observation_types=None, max_dist=50000,
        ignore_type=False, date=None, last_date=None):
    select = "*"
    order_by = "observation_type"
    params = {
        'lat': lat,
        'lon': lon,
        'max_dist': max_dist,
        'dwd_station_id': dwd_station_id,
        'wmo_station_id': wmo_station_id,
        'source_id': source_id,
        'observation_types': tuple(observation_types or ()),
        'date': date,
        'last_date': last_date,
    }
    if source_id is not None:
        if isinstance(source_id, list):
            where = "id IN %(source_id_tuple)s"
            order_by = (
                "array_position(%(source_id)s, source_id), observation_type")
            params['source_id_tuple'] = tuple(source_id)
        else:
            where = "id = %(source_id)s"
    elif dwd_station_id is not None:
        if isinstance(dwd_station_id, list):
            where = "dwd_station_id IN %(dwd_station_id_tuple)s"
            order_by = (
                "array_position(%(dwd_station_id)s, dwd_station_id::text), "
                "observation_type")
            params['dwd_station_id_tuple'] = tuple(dwd_station_id)
        else:
            where = "dwd_station_id = %(dwd_station_id)s"
    elif wmo_station_id is not None:
        if isinstance(wmo_station_id, list):
            where = "wmo_station_id IN %(wmo_station_id_tuple)s"
            order_by = (
                "array_position(%(wmo_station_id)s, wmo_station_id::text), "
                "observation_type")
            params['wmo_station_id_tuple'] = tuple(wmo_station_id)
        else:
            where = "wmo_station_id = %(wmo_station_id)s"
    elif (lat is not None and lon is not None):
        distance = """
            earth_distance(
                ll_to_earth(%(lat)s, %(lon)s),
                ll_to_earth(lat, lon)
            )
        """
        select += f", round({distance}) AS distance"
        where = f"""
            earth_box(
                ll_to_earth(%(lat)s, %(lon)s),
                %(max_dist)s
            ) @> ll_to_earth(lat, lon) AND
            {distance} < %(max_dist)s
        """
        if ignore_type:
            order_by = "distance"
        else:
            order_by += ", distance"
    else:
        raise ValueError(
            "Please supply lat/lon or dwd_station_id or wmo_station_id or "
            "source_id")
    if observation_types:
        where += " AND observation_type IN %(observation_types)s"
    if date is not None:
        where += " AND last_record >= %(date)s"
    if last_date is not None:
        where += " AND first_record <= %(last_date)s"
    sql = f"""
        SELECT {select}
        FROM sources
        WHERE {where}
        ORDER BY {order_by}
        """
    rows = fetch(sql, params)
    if not rows:
        raise LookupError("No sources match your criteria")
    return {'sources': _make_dicts(rows)}



import bz2
import csv
import datetime
import io
import json
import logging
import os
import re
import zipfile
from concurrent.futures import ProcessPoolExecutor
from contextlib import suppress

import dateutil.parser
from dateutil.tz import tzutc
from parsel import Selector

from brightsky.db import fetch
from brightsky.export import DBExporter, SYNOPExporter
from brightsky.settings import settings
from brightsky.units import (
    celsius_to_kelvin, current_observations_weather_code_to_condition,
    eighths_to_percent, hpa_to_pa, kmh_to_ms, minutes_to_seconds,
    synop_current_weather_code_to_condition,
    synop_form_of_precipitation_code_to_condition,
    synop_past_weather_code_to_condition)
from brightsky.utils import cache_path, download, dwd_id_to_wmo, wmo_id_to_dwd


class SkipRecord(Exception):
    pass


class Parser:

    DEFAULT_URL = None
    PRIORITY = 10

    exporter = DBExporter

    @property
    def logger(self):
        if not hasattr(self, '_logger'):
            self._logger = logging.getLogger(self.__class__.__name__)
        return self._logger

    def __init__(self, path=None, url=None):
        self.url = url or self.DEFAULT_URL
        self.path = path
        if not self.path and self.url:
            self.path = cache_path(self.url)
        self.downloaded_files = set()

    def download(self):
        self._download(self.url, self.path)

    def _download(self, url, path):
        self.logger.info('Downloading "%s" to "%s"', url, path)
        download_path = download(url, path)
        if download_path:
            self.downloaded_files.add(download_path)

    def should_skip(self):
        return False

    def parse(self):
        raise NotImplementedError

    def cleanup(self):
        if not settings.KEEP_DOWNLOADS:
            for path in self.downloaded_files:
                self.logger.debug("Removing '%s'", path)
                with suppress(FileNotFoundError):
                    os.remove(path)


class MOSMIXParser(Parser):

    DEFAULT_URL = (
        'https://opendata.dwd.de/weather/local_forecasts/mos/MOSMIX_S/'
        'all_stations/kml/MOSMIX_S_LATEST_240.kmz')
    PRIORITY = 20

    ELEMENTS = {
        'DD': 'wind_direction',
        'FF': 'wind_speed',
        'FX1': 'wind_gust_speed',
        'N': 'cloud_cover',
        'PPPP': 'pressure_msl',
        'RR1c': 'precipitation',
        'SunD1': 'sunshine',
        'Td': 'dew_point',
        'TTT': 'temperature',
        'VV': 'visibility',
        'ww': 'condition',
    }

    def parse(self):
        # Wrap in process as it seems to be leaking memory somewhere in lxml
        with ProcessPoolExecutor() as executor:
            f = executor.submit(MOSMIXParser._get_records, self)
            return f.result()

    @staticmethod
    def _get_records(parser):
        return list(parser._parse())

    def _parse(self):
        self.logger.info("Parsing %s", self.path)
        sel = self.get_selector()
        timestamps = self.parse_timestamps(sel)
        source = self.parse_source(sel)
        self.logger.debug(
            'Got %d timestamps for source %s', len(timestamps), source)
        for station_sel in sel.css('Placemark'):
            records = self.parse_station(station_sel, timestamps, source)
            yield from self.sanitize_records(records)

    def get_selector(self):
        with zipfile.ZipFile(self.path) as zf:
            infolist = zf.infolist()
            assert len(infolist) == 1, f'Unexpected zip content in {self.path}'
            with zf.open(infolist[0]) as f:
                sel = Selector(f.read().decode('latin1'), type='xml')
        sel.remove_namespaces()
        return sel

    def parse_timestamps(self, sel):
        return [
            dateutil.parser.parse(ts)
            for ts in sel.css('ForecastTimeSteps > TimeStep::text').extract()]

    def parse_source(self, sel):
        return ':'.join(sel.css('ProductID::text, IssueTime::text').extract())

    def parse_station(self, station_sel, timestamps, source):
        wmo_station_id = station_sel.css('name::text').extract_first()
        dwd_station_id = wmo_id_to_dwd(wmo_station_id)
        station_name = station_sel.css('description::text').extract_first()
        lon, lat, height = station_sel.css(
            'coordinates::text').extract_first().split(',')
        records = {'timestamp': timestamps}
        for element, column in self.ELEMENTS.items():
            values_str = station_sel.css(
                f'Forecast[elementName="{element}"] value::text'
            ).extract_first()
            converter = getattr(self, f'parse_{column}', float)
            records[column] = [
                None if row[0] == '-' else converter(row[0])
                for row in csv.reader(
                    re.sub(r'\s+', '\n', values_str.strip()).splitlines())
            ]
            assert len(records[column]) == len(timestamps)
        base_record = {
            'observation_type': 'forecast',
            'source': source,
            'lat': float(lat),
            'lon': float(lon),
            'height': float(height),
            'dwd_station_id': dwd_station_id,
            'wmo_station_id': wmo_station_id,
            'station_name': station_name,
        }
        # Turn dict of lists into list of dicts
        return (
            {**base_record, **dict(zip(records, row))}
            for row in zip(*records.values())
        )

    def parse_condition(self, value):
        code = int(value.split('.')[0])
        return synop_current_weather_code_to_condition(code)

    def sanitize_records(self, records):
        for r in records:
            if r['precipitation'] and r['precipitation'] < 0:
                self.logger.warning(
                    "Ignoring negative precipitation value: %s", r)
                r['precipitation'] = None
            if r['wind_direction'] and r['wind_direction'] > 360:
                self.logger.warning(
                    "Fixing out-of-bounds wind direction: %s", r)
                r['wind_direction'] -= 360
            yield r


class SYNOPParser(Parser):

    PRIORITY = 30

    exporter = SYNOPExporter

    mandatory_fields = [
        'wmo_station_id', 'station_name', 'lat', 'lon', 'height', 'timestamp']

    elements = {
        'cloudCoverTotal': 'cloud_cover',
        'heightOfStationGroundAboveMeanSeaLevel': 'height',
        'latitude': 'lat',
        'longitude': 'lon',
        'meteorologicalOpticalRange': 'visibility',
        'pressureReducedToMeanSeaLevel': 'pressure_msl',
        'stationOrSiteName': 'station_name',
    }
    height_field = 'heightOfSensorAboveLocalGroundOrDeckOfMarinePlatform'
    height = 2
    height_elements = {
        'airTemperature': 'temperature',
        'dewpointTemperature': 'dew_point',
        'relativeHumidity': 'relative_humidity',
    }
    time_period_field = 'timePeriod'
    time_periods = (-10, -30, -60)
    time_period_elements = {
        'windDirection': 'wind_direction',
        'windSpeed': 'wind_speed',
        'maximumWindGustDirection': 'wind_gust_direction',
        'maximumWindGustSpeed': 'wind_gust_speed',
        'totalPrecipitationOrTotalWaterEquivalent': 'precipitation',
        'totalSunshine': 'sunshine',
    }

    def parse(self):
        with bz2.open(self.path) as f:
            if next(f).startswith(b'no messages found'):
                return
            f.seek(0)
            message_blocks = json.load(f)['messages']
        for block in message_blocks:
            for message in block[-1]:
                with suppress(SkipRecord):
                    record = self.parse_message(message)
                    self.sanitize_record(record)
                    yield record

    def parse_message(self, message):
        record = {
            'observation_type': 'synop',
        }
        self.parse_tree(record, message)
        is_complete = all(
            record.get(field) is not None for field in self.mandatory_fields)
        if not is_complete:
            self.logger.error("Skipping incomplete record: %s", record)
            raise SkipRecord
        return record

    def parse_tree(self, record, message, base=None):
        data = {} if base is None else base.copy()
        for block in message:
            if isinstance(block, dict):
                key = block['key']
                value = block['value']
                data[key] = value
                if field := self.elements.get(key):
                    record[field] = value
                elif field := self.height_elements.get(key):
                    if data[self.height_field] == self.height:
                        record[field] = value
                elif field := self.time_period_elements.get(key):
                    time_period = data[self.time_period_field]
                    if time_period in self.time_periods:
                        if field == 'sunshine' and value:
                            value *= 60
                        record[field + f'_{-time_period}'] = value
                elif parse_method := getattr(self, f'parse_{key}', None):
                    parse_method(record, data, value)
            else:
                self.parse_tree(record, block, base=data)

    def parse_minute(self, record, data, value):
        parts = ['year', 'month', 'day', 'hour', 'minute']
        if any(data[part] is None for part in parts):
            raise SkipRecord
        record['timestamp'] = datetime.datetime(
            data['year'], data['month'], data['day'], data['hour'],
            data['minute'], tzinfo=tzutc())

    def parse_stationNumber(self, record, data, value):
        if data['stationNumber']:
            wmo_id = f"{data['blockNumber']}{data['stationNumber']:03d}"
        else:
            wmo_id = data['shortStationName']
        record['wmo_station_id'] = wmo_id
        record['dwd_station_id'] = wmo_id_to_dwd(wmo_id)

    def parse_presentWeather(self, record, data, value):
        if record.get('timePeriod'):
            return
        condition = synop_current_weather_code_to_condition(value)
        # Don't overwrite any condition from pastWeather1 with None, but do
        # prioritize presentWeather if it's not None
        if condition:
            record['condition'] = condition

    def parse_pastWeather1(self, record, data, value):
        if value and not record.get('condition'):
            record['condition'] = synop_past_weather_code_to_condition(value)

    def sanitize_record(self, record):
        for field in list(record):
            if field.startswith('precipitation_') and (record[field] or 0) < 0:
                record[field] = None
        if (record.get('cloud_cover') or 0) > 100:
            record['cloud_cover'] = None


class CurrentObservationsParser(Parser):

    PRIORITY = 30

    ELEMENTS = {
        'cloud_cover_total': 'cloud_cover',
        'dew_point_temperature_at_2_meter_above_ground': 'dew_point',
        'dry_bulb_temperature_at_2_meter_above_ground': 'temperature',
        'horizontal_visibility': 'visibility',
        'maximum_wind_speed_last_hour': 'wind_gust_speed',
        'mean_wind_direction_during_last_10 min_at_10_meters_above_ground': (
            'wind_direction'),
        'mean_wind_speed_during last_10_min_at_10_meters_above_ground': (
            'wind_speed'),
        'precipitation_amount_last_hour': 'precipitation',
        'present_weather': 'condition',
        'pressure_reduced_to_mean_sea_level': 'pressure_msl',
        'relative_humidity': 'relative_humidity',
        'total_time_of_sunshine_during_last_hour': 'sunshine',
    }
    DATE_COLUMN = 'surface observations'
    HOUR_COLUMN = 'Parameter description'

    CONVERTERS = {
        'condition': current_observations_weather_code_to_condition,
        'dew_point': celsius_to_kelvin,
        'pressure_msl': hpa_to_pa,
        'sunshine': minutes_to_seconds,
        'temperature': celsius_to_kelvin,
        'wind_speed': kmh_to_ms,
        'wind_gust_speed': kmh_to_ms,
    }
    IGNORED_VALUES = {
        'cloud_cover': ['112', '113', '126'],
        'relative_humidity': ['101', '107'],
    }

    def should_skip(self):
        return self.path.endswith(tuple(
            f'{station:_<5}-BEOB.csv'
            for station in settings.IGNORED_CURRENT_OBSERVATIONS_STATIONS))

    def parse(self, lat=None, lon=None, height=None, station_name=None):
        with open(self.path) as f:
            reader = csv.DictReader(f, delimiter=';')
            wmo_station_id = next(reader)[self.DATE_COLUMN].rstrip('_')
            dwd_station_id = wmo_id_to_dwd(wmo_station_id)
            if any(x is None for x in (lat, lon, height, station_name)):
                lat, lon, height, station_name = self.load_location(
                    wmo_station_id)
            # Skip row with German header titles
            next(reader)
            for row in reader:
                yield {
                    'observation_type': 'current',
                    'lat': lat,
                    'lon': lon,
                    'height': height,
                    'dwd_station_id': dwd_station_id,
                    'wmo_station_id': wmo_station_id,
                    'station_name': station_name,
                    **self.parse_row(row)
                }

    def parse_row(self, row):
        record = {
            element: (
                None
                if row[column] == '---'
                or row[column] in self.IGNORED_VALUES.get(element, [])
                else float(row[column].replace(',', '.')))
            for column, element in self.ELEMENTS.items()
        }
        record['timestamp'] = datetime.datetime.strptime(
            f'{row[self.DATE_COLUMN]} {row[self.HOUR_COLUMN]}',
            '%d.%m.%y %H:%M'
        ).replace(tzinfo=tzutc())
        self.convert_units(record)
        return record

    def convert_units(self, record):
        for element, converter in self.CONVERTERS.items():
            if record[element] is not None:
                record[element] = converter(record[element])

    def load_location(self, wmo_station_id):
        rows = fetch(
            """
            SELECT lat, lon, height, station_name
            FROM sources
            WHERE wmo_station_id = %s
            ORDER BY observation_type DESC, id DESC
            LIMIT 1
            """,
            (wmo_station_id,),
        )
        if not rows:
            raise ValueError(
                f'Unable to find location for WMO station {wmo_station_id}')
        return rows[0]


class ObservationsParser(Parser):

    elements = {}
    converters = {}
    ignored_values = {}

    def should_skip(self):
        if (m := re.search(r'_(\d{8})_(\d{8})_hist\.zip$', str(self.path))):
            end_date = datetime.datetime.strptime(
                m.group(2), '%Y%m%d').replace(tzinfo=tzutc())
            if end_date < settings.MIN_DATE:
                return True
            if settings.MAX_DATE:
                start_date = datetime.datetime.strptime(
                    m.group(1), '%Y%m%d').replace(tzinfo=tzutc())
                if start_date > settings.MAX_DATE:
                    return True
        return False

    def parse(self):
        with zipfile.ZipFile(self.path) as zf:
            dwd_station_id = self.parse_station_id(zf)
            wmo_station_id = dwd_id_to_wmo(dwd_station_id)
            observation_type = self.parse_observation_type()
            lat_lon_history = self.parse_lat_lon_history(zf, dwd_station_id)
            for record in self.parse_records(zf, lat_lon_history):
                yield {
                    'observation_type': observation_type,
                    'dwd_station_id': dwd_station_id,
                    'wmo_station_id': wmo_station_id,
                    **record
                }

    def parse_station_id(self, zf):
        for filename in zf.namelist():
            if (m := re.match(r'Metadaten_Geographie_(\d+)\.txt', filename)):
                return m.group(1)
        raise ValueError(f"Unable to parse station ID for {self.path}")

    def parse_observation_type(self):
        filename = os.path.basename(self.path)
        if filename.endswith('_akt.zip'):
            return 'recent'
        elif filename.endswith('_hist.zip'):
            return 'historical'
        raise ValueError(
            f'Unable to determine observation type from path "{self.path}"')

    def parse_lat_lon_history(self, zf, dwd_station_id):
        with zf.open(f'Metadaten_Geographie_{dwd_station_id}.txt') as f:
            reader = csv.DictReader(
                io.TextIOWrapper(f, encoding='latin1'),
                delimiter=';')
            history = {}
            for row in reader:
                date_from = datetime.datetime.strptime(
                    row['von_datum'].strip(), '%Y%m%d'
                ).replace(tzinfo=tzutc())
                history[date_from] = (
                    float(row['Geogr.Breite']),
                    float(row['Geogr.Laenge']),
                    float(row['Stationshoehe']),
                    row['Stationsname'])
            return history

    def parse_records(self, zf, lat_lon_history):
        product_filenames = [
            fn for fn in zf.namelist() if fn.startswith('produkt_')]
        assert len(product_filenames) == 1, "Unexpected product count"
        filename = product_filenames[0]
        with zf.open(filename) as f:
            reader = csv.DictReader(
                io.TextIOWrapper(f, encoding='latin1'),
                delimiter=';')
            yield from self.parse_reader(filename, reader, lat_lon_history)

    def parse_reader(self, filename, reader, lat_lon_history):
        for row in reader:
            timestamp = datetime.datetime.strptime(
                row['MESS_DATUM'], '%Y%m%d%H').replace(tzinfo=tzutc())
            if self._skip_timestamp(timestamp):
                continue
            lat, lon, height, station_name = self._station_params(
                timestamp, lat_lon_history)
            yield {
                'source': f'Observations:Recent:{filename}',
                'lat': lat,
                'lon': lon,
                'height': height,
                'station_name': station_name,
                'timestamp': timestamp,
                **self.parse_elements(row, lat, lon, height),
            }

    def _skip_timestamp(self, timestamp):
        return (
            timestamp < settings.MIN_DATE or
            (settings.MAX_DATE and timestamp > settings.MAX_DATE))

    def _station_params(self, timestamp, lat_lon_history):
        info = None
        for date, lat_lon_height_name in lat_lon_history.items():
            if date > timestamp:
                break
            info = lat_lon_height_name
        return info

    def parse_elements(self, row, lat, lon, height):
        elements = {
            element: (
                float(row[element_key])
                if row[element_key].strip() != '-999'
                and row[element_key].strip() not in self.ignored_values.get(
                    element, [])
                else None)
            for element, element_key in self.elements.items()
        }
        for element, converter in self.converters.items():
            if elements[element] is not None:
                elements[element] = converter(elements[element])
        return elements


class CloudCoverObservationsParser(ObservationsParser):

    elements = {
        'cloud_cover': ' V_N',
    }
    ignored_values = {
        'cloud_cover': ['-1', '9'],
    }
    converters = {
        'cloud_cover': eighths_to_percent,
    }


class DewPointObservationsParser(ObservationsParser):

    elements = {
        'dew_point': '  TD',
    }
    converters = {
        'dew_point': celsius_to_kelvin,
    }


class TemperatureObservationsParser(ObservationsParser):

    elements = {
        'relative_humidity': 'RF_TU',
        'temperature': 'TT_TU',
    }
    converters = {
        'temperature': celsius_to_kelvin,
    }


class PrecipitationObservationsParser(ObservationsParser):

    elements = {
        'precipitation': '  R1',
        'condition': 'WRTR',
    }
    converters = {
        'condition': synop_form_of_precipitation_code_to_condition,
    }

    def parse_reader(self, filename, reader, lat_lon_history):
        # XXX: WRTR is missing every third hour, we fill it up from the
        #      previous or next row where sensible
        rows = list(reader)
        for i, row in enumerate(rows):
            if row['WRTR'] != '-999':
                continue
            elif row['RS_IND'].strip() == '0':
                row['WRTR'] = '0'
            elif i > 0 and rows[i-1]['RS_IND'].strip() == '1':
                row['WRTR'] = rows[i-1]['WRTR']
            elif i + 1 < len(rows) and rows[i+1]['RS_IND'].strip() == '1':
                row['WRTR'] = rows[i+1]['WRTR']
            else:
                row['WRTR'] = '9'
        yield from super().parse_reader(filename, rows, lat_lon_history)


class VisibilityObservationsParser(ObservationsParser):

    elements = {
        'visibility': 'V_VV',
    }
    converters = {
        'visibility': int,
    }


class WindObservationsParser(ObservationsParser):

    elements = {
        'wind_speed': '   F',
        'wind_direction': '   D',
    }
    converters = {
        'wind_direction': int,
    }
    ignored_values = {
        'wind_direction': ['990'],
    }


class WindGustsObservationsParser(ObservationsParser):

    META_DATA_URL = (
        'https://opendata.dwd.de/climate_environment/CDC/observations_germany/'
        'climate/10_minutes/extreme_wind/meta_data/'
        'Meta_Daten_zehn_min_fx_{dwd_station_id}.zip')

    elements = {
        'wind_gust_direction': 'DX_10',
        'wind_gust_speed': 'FX_10',
    }

    def __init__(self, *args, meta_path=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.meta_path = meta_path

    def download(self):
        super().download()
        with zipfile.ZipFile(self.path) as zf:
            dwd_station_id = self.parse_station_id(zf)
        meta_data_url = self.META_DATA_URL.format(
            dwd_station_id=dwd_station_id)
        if not self.meta_path:
            self.meta_path = cache_path(meta_data_url)
        self._download(meta_data_url, self.meta_path)

    def parse_station_id(self, zf):
        for filename in zf.namelist():
            if (m := re.match(r'produkt_.*_(\d+)\.txt', filename)):
                return m.group(1)
        raise ValueError(f"Unable to parse station ID for {self.path}")

    def parse_lat_lon_history(self, zf, dwd_station_id):
        with zipfile.ZipFile(self.meta_path) as meta_zf:
            return super().parse_lat_lon_history(meta_zf, dwd_station_id)

    def parse_reader(self, filename, reader, lat_lon_history):
        hour_values = []
        # First row is at :00, which we will already have filled up with
        # the last :50 entry of another file (see below)
        next(reader)
        for row in reader:
            timestamp = datetime.datetime.strptime(
                row['MESS_DATUM'], '%Y%m%d%H%M').replace(tzinfo=tzutc())
            if self._skip_timestamp(timestamp + datetime.timedelta(hours=1)):
                continue
            # Should this be refactored into a base class we will need to
            # properly parse the station parameters and pass them
            values = self.parse_elements(row, None, None, None)
            if values['wind_gust_speed']:
                hour_values.append(values)
            if timestamp.minute == 0:
                yield self._make_record(
                    timestamp, hour_values, filename, lat_lon_history)
                hour_values.clear()
        observation_type = self.parse_observation_type()
        if observation_type == 'historical' and timestamp.minute == 50:
            # Not 100 % accurate but better than taking only the :00 value of
            # another file. For observation_type 'recent', we'll get a proper
            # midnight value from the 'current' observation
            yield self._make_record(
                timestamp + datetime.timedelta(minutes=10),
                hour_values, filename, lat_lon_history)

    def _make_record(self, timestamp, hour_values, filename, lat_lon_history):
        lat, lon, height, station_name = self._station_params(
            timestamp, lat_lon_history)
        if hour_values:
            max_value = max(hour_values, key=lambda v: v['wind_gust_speed'])
            direction = max_value['wind_gust_direction']
            speed = max_value['wind_gust_speed']
        else:
            direction = None
            speed = None
        return {
            'source': f'Observations:Recent:{filename}',
            'lat': lat,
            'lon': lon,
            'height': height,
            'station_name': station_name,
            'timestamp': timestamp,
            'wind_gust_direction': direction,
            'wind_gust_speed': speed,
        }


class SunshineObservationsParser(ObservationsParser):

    elements = {
        'sunshine': 'SD_SO',
    }
    converters = {
        'sunshine': minutes_to_seconds,
    }


class PressureObservationsParser(ObservationsParser):

    elements = {
        'pressure_msl': '   P',
        'pressure_station': '  P0',
    }
    converters = {
        'pressure_msl': hpa_to_pa,
        'pressure_station': hpa_to_pa,
    }

    def parse_elements(self, row, lat, lon, height):
        elements = super().parse_elements(row, lat, lon, height)
        if not elements['pressure_msl'] and elements['pressure_station']:
            # Some stations do not record reduced pressure, but do record
            # pressure at station height. We can approximate the pressure at
            # mean sea level through the barometric formula. The error of this
            # approximation could be reduced if we had the current temperature.
            elements['pressure_msl'] = int(round(
                elements['pressure_station']
                * (1 - .0065 * height / 288.15) ** -5.255,
                -1))
        del elements['pressure_station']
        return elements


def get_parser(filename):
    parsers = {
        r'MOSMIX_S_LATEST_240\.kmz$': MOSMIXParser,
        r'Z__C_EDZW_\d+_.*\.json\.bz2$': SYNOPParser,
        r'\w{5}-BEOB\.csv$': CurrentObservationsParser,
        'stundenwerte_FF_': WindObservationsParser,
        'stundenwerte_N_': CloudCoverObservationsParser,
        'stundenwerte_P0_': PressureObservationsParser,
        'stundenwerte_RR_': PrecipitationObservationsParser,
        'stundenwerte_SD_': SunshineObservationsParser,
        'stundenwerte_TD_': DewPointObservationsParser,
        'stundenwerte_TU_': TemperatureObservationsParser,
        'stundenwerte_VV_': VisibilityObservationsParser,
        '10minutenwerte_extrema_wind_': WindGustsObservationsParser,
    }
    for pattern, parser in parsers.items():
        if re.match(pattern, filename):
            return parser



import datetime
import logging
import threading
import time
import os
from contextlib import suppress
from functools import lru_cache

import coloredlogs
import dateutil.parser
from astral import Observer
from astral.sun import daylight
from dateutil.tz import tzlocal, tzutc
from parsel import Selector

import requests


logger = logging.getLogger(__name__)


USER_AGENT = 'Bright Sky / https://brightsky.dev/'


def configure_logging():
    log_fmt = '%(asctime)s %(name)s %(levelname)s  %(message)s'
    coloredlogs.install(level=logging.DEBUG, fmt=log_fmt)
    # Disable some third-party noise
    logging.getLogger('huey').setLevel(logging.INFO)
    logging.getLogger('urllib3').setLevel(logging.WARNING)


def load_dotenv(path='.env'):
    with suppress(FileNotFoundError):
        with open(path) as f:
            for line in f:
                if line.strip() and not line.strip().startswith('#'):
                    key, val = line.strip().split('=', 1)
                    os.environ.setdefault(key, val)


def download(url, path):
    """
    Download a resource from `url` to `path`, unless the current version
    already lives at `path`.
    """
    if os.path.isfile(path):
        last_modified_local = datetime.datetime.utcfromtimestamp(
            os.path.getmtime(path)).replace(tzinfo=tzutc())
        resp = requests.head(url)
        resp.raise_for_status()
        last_modified_api = dateutil.parser.parse(
            resp.headers['Last-Modified'])
        is_outdated = last_modified_api > last_modified_local
        if not is_outdated:
            logger.debug(
                '%s is already the newest version, skipping download from %s',
                path, url)
            return
    resp = requests.get(url, headers={'User-Agent': USER_AGENT})
    resp.raise_for_status()
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, 'wb') as f:
        f.write(resp.content)
    last_modified = dateutil.parser.parse(
        resp.headers['Last-Modified']).timestamp()
    os.utime(path, (last_modified, last_modified))
    return path


def cache_path(url):
    dirname = os.path.join(os.getcwd(), '.cache/brightsky')
    filename = os.path.basename(url)
    return os.path.join(dirname, filename)


def dwd_fingerprint(path):
    """Return file attributes in same format as DWD server index pages"""
    last_modified = datetime.datetime.fromtimestamp(
        os.path.getmtime(path)
    ).replace(
        second=0,
        tzinfo=tzlocal()
    ).astimezone(tzutc())
    return {
        'last_modified': last_modified,
        'file_size': os.path.getsize(path),
    }


def parse_date(date_str):
    d = dateutil.parser.parse(date_str)
    if not d.tzinfo:
        d.replace(tzinfo=tzutc())
    return d


@lru_cache
def sunrise_sunset(lat, lon, date):
    return daylight(Observer(lat, lon), date)


class StationIDConverter:

    STATION_LIST_URL = (
        'https://www.dwd.de/DE/leistungen/klimadatendeutschland/statliste/'
        'statlex_html.html?view=nasPublication')
    STATION_TYPES = ['SY', 'MN']
    UPDATE_INTERVAL = 86400

    update_lock = threading.Lock()

    def __init__(self):
        self.last_update = 0
        self.dwd_to_wmo = {}
        self.wmo_to_dwd = {}

    def update(self, force=False):
        with self.update_lock:
            is_recent = time.time() - self.last_update < self.UPDATE_INTERVAL
            if not force and is_recent:
                return
            logger.info("Updating station ID maps")
            resp = requests.get(
                self.STATION_LIST_URL, headers={'User-Agent': USER_AGENT})
            resp.raise_for_status()
            self.parse_station_list(resp.text)

    def parse_station_list(self, html):
        sel = Selector(html)
        station_rows = []
        for station_type in self.STATION_TYPES:
            station_rows.extend(sel.xpath(
                f'//tr[td[3][text() = "{station_type}"]]'))
        assert station_rows, "No synoptic stations"
        self.dwd_to_wmo.clear()
        self.wmo_to_dwd.clear()
        for row in station_rows:
            values = row.css('td::text').extract()
            dwd_id = values[1].zfill(5)
            wmo_id = values[3]
            self.dwd_to_wmo[dwd_id] = wmo_id
            self.wmo_to_dwd[wmo_id] = dwd_id
        self.last_update = time.time()
        logger.info("Parsed %d station ID mappings", len(station_rows))

    def convert_to_wmo(self, dwd_id):
        self.update()
        return self.dwd_to_wmo.get(dwd_id)

    def convert_to_dwd(self, wmo_id):
        self.update()
        return self.wmo_to_dwd.get(wmo_id)


_converter = StationIDConverter()
dwd_id_to_wmo = _converter.convert_to_wmo
wmo_id_to_dwd = _converter.convert_to_dwd



import logging
import os

from brightsky.db import get_connection
from brightsky.parsers import get_parser
from brightsky.polling import DWDPoller
from brightsky.utils import dwd_fingerprint


logger = logging.getLogger('brightsky')


def parse(path=None, url=None, export=False):
    if not path and not url:
        raise ValueError('Please provide either path or url')
    parser_cls = get_parser(os.path.basename(path or url))
    parser = parser_cls(path=path, url=url)
    if url:
        parser.download()
        fingerprint = {
            'url': url,
            **dwd_fingerprint(parser.path),
        }
    else:
        fingerprint = None
    records = list(parser.parse())
    parser.cleanup()
    if export:
        exporter = parser.exporter()
        exporter.export(records, fingerprint=fingerprint)
    return records


def poll(enqueue=False):
    updated_files = DWDPoller().poll()
    if enqueue:
        from brightsky.worker import huey, process
        if (expired_locks := huey.expire_locks(1800)):
            logger.warning(
                'Removed expired locks: %s', ', '.join(expired_locks))
        pending_urls = [
            t.args[0] for t in huey.pending() if t.name == 'process']
        enqueued = 0
        for updated_file in updated_files:
            url = updated_file['url']
            if url in pending_urls:
                logger.debug('Skipping "%s": already queued', url)
                continue
            elif huey.is_locked(url):
                logger.debug('Skipping "%s": already running', url)
                continue
            logger.debug('Enqueueing "%s"', url)
            parser_cls = get_parser(os.path.basename(url))
            process(url, priority=parser_cls.PRIORITY)
            enqueued += 1
        logger.info(
            'Enqueued %d updated files for processing. Queue size: %d',
            enqueued, enqueued + len(pending_urls))
    return updated_files


def clean():
    expiry_intervals = {
        'weather': {
            'forecast': '3 hours',
            'current': '48 hours',
        },
        'synop': {
            'synop': '30 hours',
        },
    }
    parsed_files_expiry_intervals = {
        '%/Z__C_EDZW_%': '1 week',
    }
    with get_connection() as conn:
        with conn.cursor() as cur:
            # XXX: This assumes that the DWD will upload 'historical' records
            #      for ALL weather parameters at the same time. If this turns
            #      out to be false we should merge the 'historical' and
            #      'recent' observation types into a single one, otherwise we
            #      will have periods where 'historical' records miss certain
            #      weather parameters, and we cannot fall back to the 'recent'
            #      records as they are gone.
            logger.info("Deleting obsolete 'recent' weather records")
            cur.execute(
                """
                SELECT
                    s_recent.id,
                    s_historical.last_record AS threshold
                FROM sources s_recent
                JOIN sources s_historical ON (
                    s_recent.wmo_station_id = s_historical.wmo_station_id AND
                    s_recent.dwd_station_id = s_historical.dwd_station_id)
                WHERE
                    s_recent.observation_type = 'recent' AND
                    s_historical.observation_type = 'historical' AND
                    s_recent.first_record < s_historical.last_record
                """)
            for row in cur.fetchall():
                logger.debug(
                    "Deleting records for source %d prior to %s",
                    row['id'], row['threshold'])
                cur.execute(
                    """
                    DELETE FROM weather
                    WHERE source_id = %s AND timestamp < %s
                    """,
                    (row['id'], row['threshold']))
            conn.commit()
            logger.info(
                'Deleting expired weather records: %s', expiry_intervals)
            for table, table_expires in expiry_intervals.items():
                for observation_type, interval in table_expires.items():
                    cur.execute(
                        f"""
                        DELETE FROM {table} WHERE
                            source_id IN (
                                SELECT id FROM sources
                                WHERE observation_type = %s) AND
                            timestamp < current_timestamp - %s::interval;
                        """,
                        (observation_type, interval),
                    )
                    conn.commit()
                    if cur.rowcount:
                        logger.info(
                            'Deleted %d outdated %s weather records from %s',
                            cur.rowcount, observation_type, table)
                cur.execute(
                    f"""
                    UPDATE sources SET
                      first_record = record_range.first_record,
                      last_record = record_range.last_record
                    FROM (
                      SELECT
                        source_id,
                        MIN(timestamp) AS first_record,
                        MAX(timestamp) AS last_record
                      FROM {table}
                      GROUP BY source_id
                    ) AS record_range
                    WHERE sources.id = record_range.source_id;
                    """)
                conn.commit()
            logger.info(
                'Deleting expired parsed files: %s',
                parsed_files_expiry_intervals)
            for filename, interval in parsed_files_expiry_intervals.items():
                cur.execute(
                    """
                    DELETE FROM parsed_files WHERE
                        url LIKE %s AND
                        parsed_at < current_timestamp - %s::interval;
                    """,
                    (filename, interval))
                conn.commit()
                if cur.rowcount:
                    logger.info(
                        'Deleted %d outdated parsed files for pattern "%s"',
                        cur.rowcount, filename)



import datetime
import os

from dateutil.tz import tzutc


CORS_ALLOWED_ORIGINS = []
DATABASE_URL = 'postgres://localhost'
ICON_CLOUDY_THRESHOLD = 80
ICON_PARTLY_CLOUDY_THRESHOLD = 25
ICON_RAIN_THRESHOLD = 0.5
ICON_WIND_THRESHOLD = 10.8
IGNORED_CURRENT_OBSERVATIONS_STATIONS = ['K386']
KEEP_DOWNLOADS = False
MIN_DATE = datetime.datetime(2010, 1, 1, tzinfo=tzutc())
MAX_DATE = None
POLLING_CRONTAB_MINUTE = '*'
REDIS_URL = 'redis://localhost'


def _make_bool(bool_str):
    return bool_str == '1'


def _make_date(date_str):
    return datetime.datetime.fromisoformat(date_str).replace(tzinfo=tzutc())


def _make_list(list_str, separator=','):
    return list_str.split(separator)


_SETTING_PARSERS = {
    'MAX_DATE': _make_date,

    bool: _make_bool,
    datetime.datetime: _make_date,
    float: float,
    list: _make_list,
}


class Settings(dict):
    """A dictionary that makes its keys available as attributes"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.loaded = False

    def load(self):
        for k, v in globals().items():
            if k.isupper() and not k.startswith('_'):
                self[k] = v
        for k, v in os.environ.items():
            if k.startswith('BRIGHTSKY_') and k.isupper():
                setting_name = k.split('_', 1)[1]
                setting_type = type(self.get(setting_name))
                setting_parser = _SETTING_PARSERS.get(
                    setting_name, _SETTING_PARSERS.get(setting_type))
                if setting_parser:
                    v = setting_parser(v)
                self[setting_name] = v

    def __getattr__(self, name):
        if not self.loaded:
            self.load()
            self.loaded = True
        return self[name]


settings = Settings()



import functools
import logging
from threading import Lock

from psycopg2 import sql
from psycopg2.extras import execute_values

from brightsky.db import get_connection


logger = logging.getLogger(__name__)


class DBExporter:

    # The ON CONFLICT clause won't change anything in most cases, but it
    # ensures that the row is always returned (so that we can build the source
    # map).
    UPDATE_SOURCES_STMT = """
        INSERT INTO sources (
            observation_type, lat, lon, height, dwd_station_id, wmo_station_id,
            station_name, first_record, last_record)
        VALUES %s
        ON CONFLICT
            ON CONSTRAINT weather_source_key DO UPDATE SET
                dwd_station_id = EXCLUDED.dwd_station_id,
                wmo_station_id = EXCLUDED.wmo_station_id,
                station_name = EXCLUDED.station_name,
                first_record = LEAST(
                    sources.first_record, EXCLUDED.first_record),
                last_record = GREATEST(
                    sources.last_record, EXCLUDED.last_record)
        RETURNING id;
    """
    UPDATE_SOURCES_CLEANUP = """
        SELECT setval('sources_id_seq', (SELECT max(id) FROM sources));
    """
    WEATHER_TABLE = 'weather'
    UPDATE_WEATHER_STMT = sql.SQL("""
        INSERT INTO {weather_table} (timestamp, source_id, {fields})
        VALUES %s
        ON CONFLICT
            ON CONSTRAINT {constraint} DO UPDATE SET
                {conflict_updates};
    """)
    UPDATE_WEATHER_CONFLICT_UPDATE = '{field} = EXCLUDED.{field}'
    UPDATE_WEATHER_CLEANUP = None
    SOURCE_FIELDS = [
        'observation_type', 'lat', 'lon', 'height', 'dwd_station_id',
        'wmo_station_id', 'station_name']
    ELEMENT_FIELDS = [
        'cloud_cover', 'condition', 'dew_point', 'precipitation',
        'pressure_msl', 'relative_humidity', 'sunshine', 'temperature',
        'visibility', 'wind_direction', 'wind_speed', 'wind_gust_direction',
        'wind_gust_speed']

    sources_update_lock = Lock()

    def export(self, records, fingerprint=None):
        records = self.prepare_records(records)
        sources = self.prepare_sources(records)
        with get_connection() as conn:
            source_map = self.update_sources(conn, sources)
            self.update_weather(conn, source_map, records)
            if fingerprint:
                self.update_parsed_files(conn, fingerprint)

    def prepare_records(self, records):
        return records

    def prepare_sources(self, records):
        sources = {}
        for r in records:
            r['source'] = tuple(r[field] for field in self.SOURCE_FIELDS)
            source = sources.setdefault(
                r['source'],
                {field: r[field] for field in self.SOURCE_FIELDS})
            if 'first_record' in source:
                source['first_record'] = min(
                    source['first_record'], r['timestamp'])
                source['last_record'] = max(
                    source['last_record'], r['timestamp'])
            else:
                source['first_record'] = r['timestamp']
                source['last_record'] = r['timestamp']
        return sources

    def update_sources(self, conn, sources):
        extra_fields = ['first_record', 'last_record']
        fields = ', '.join(
            f'%({field})s' for field in self.SOURCE_FIELDS + extra_fields)
        with self.sources_update_lock:
            with conn.cursor() as cur:
                rows = execute_values(
                    cur, self.UPDATE_SOURCES_STMT, sources.values(),
                    template=f'({fields})', fetch=True)
                cur.execute(self.UPDATE_SOURCES_CLEANUP)
                conn.commit()
        return {
            source_key: row['id']
            for row, source_key in zip(rows, sources)
        }

    def update_weather(self, conn, source_map, records):
        for r in records:
            r['source_id'] = source_map[r['source']]
        for fields, records in self.make_batches(records).items():
            logger.info(
                "Exporting %d records with fields %s",
                len(records), tuple(fields))
            stmt = self.UPDATE_WEATHER_STMT.format(
                weather_table=sql.Identifier(self.WEATHER_TABLE),
                constraint=sql.Identifier(f'{self.WEATHER_TABLE}_key'),
                fields=sql.SQL(', ').join(
                    sql.Identifier(f) for f in fields),
                conflict_updates=sql.SQL(', ').join(
                    sql.SQL(self.UPDATE_WEATHER_CONFLICT_UPDATE).format(
                        field=sql.Identifier(f),
                        weather_table=sql.Identifier(self.WEATHER_TABLE))
                    for f in fields),
            )
            template = sql.SQL(
                "(%(timestamp)s, %(source_id)s, {values})"
            ).format(
                values=sql.SQL(', ').join(
                    sql.Placeholder(f) for f in fields),
            )
            with conn.cursor() as cur:
                execute_values(cur, stmt, records, template, page_size=1000)
        if self.UPDATE_WEATHER_CLEANUP:
            with conn.cursor() as cur:
                cur.execute(self.UPDATE_WEATHER_CLEANUP)

    def make_batches(self, records):
        batches = {}
        for record in records:
            fields = frozenset(f for f in self.ELEMENT_FIELDS if f in record)
            assert fields, "Got record without element fields"
            batch = batches.setdefault(fields, [])
            batch.append(record)
        return batches

    def update_parsed_files(self, conn, fingerprint):
        with conn.cursor() as cur:
            cur.execute(
                """
                INSERT INTO parsed_files (
                    url, last_modified, file_size, parsed_at
                )
                VALUES (
                    %(url)s, %(last_modified)s, %(file_size)s,
                    current_timestamp
                )
                ON CONFLICT
                    ON CONSTRAINT parsed_files_pkey DO UPDATE SET
                        last_modified = %(last_modified)s,
                        file_size = %(file_size)s,
                        parsed_at = current_timestamp;
                """,
                fingerprint)


class SYNOPExporter(DBExporter):

    WEATHER_TABLE = 'synop'
    UPDATE_WEATHER_CONFLICT_UPDATE = (
        '{field} = COALESCE(EXCLUDED.{field}, {weather_table}.{field})')
    UPDATE_WEATHER_CLEANUP = (
        'REFRESH MATERIALIZED VIEW CONCURRENTLY current_weather')

    ELEMENT_FIELDS = [
        'cloud_cover', 'condition', 'dew_point', 'precipitation_10',
        'precipitation_30', 'precipitation_60', 'pressure_msl',
        'relative_humidity', 'sunshine_10', 'sunshine_30', 'sunshine_60',
        'temperature', 'visibility', 'wind_direction_10', 'wind_direction_30',
        'wind_direction_60', 'wind_speed_10', 'wind_speed_30', 'wind_speed_60',
        'wind_gust_direction_10', 'wind_gust_direction_30',
        'wind_gust_direction_60', 'wind_gust_speed_10', 'wind_gust_speed_30',
        'wind_gust_speed_60']

    synop_update_lock = Lock()

    def prepare_records(self, records):
        # Merge records for same source and timestamp (otherwise we will run
        # into trouble with our ON CONFLICT DO UPDATE as we cannot touch the
        # same row twice in the same command).
        records_by_key = {}
        for r in records:
            key = (r['timestamp'], r['wmo_station_id'])
            records_by_key.setdefault(key, []).append(r)
        return [
            functools.reduce(self._update_where_none, records)
            for records in records_by_key.values()]

    def _update_where_none(self, base, update):
        for k, v in update.items():
            if not base.get(k):
                base[k] = v
        return base

    def update_weather(self, *args, **kwargs):
        with self.synop_update_lock:
            super().update_weather(*args, **kwargs)



import datetime
import json
from multiprocessing import cpu_count

import click
from dateutil.tz import tzutc
from huey.consumer_options import ConsumerConfig

from brightsky import db, tasks, query
from brightsky.utils import parse_date


def dump_records(it):
    for record in it:
        print(json.dumps(record, default=str))


def migrate_callback(ctx, param, value):
    if value:
        db.migrate()


def parse_date_arg(ctx, param, value):
    if not value:
        return
    return parse_date(value)


@click.group()
@click.option(
    '--migrate', help='Migrate database before running command',
    is_flag=True, is_eager=True, expose_value=False, callback=migrate_callback)
def cli():
    pass


@cli.command(help='Apply all pending database migrations')
def migrate():
    db.migrate()


@cli.command(help='Parse a forecast or observations file')
@click.option('--path', help='Local file path to observations file')
@click.option('--url', help='URL of observations file')
@click.option(
    '--export/--no-export', default=False,
    help='Export parsed records to database')
def parse(path, url, export):
    if not path and not url:
        raise click.ClickException('Please provide either --path or --url')
    records = tasks.parse(path=path, url=url, export=export)
    if not export:
        dump_records(records)


@cli.command(help='Detect updated files on DWD Open Data Server')
@click.option(
    '--enqueue/--no-enqueue', default=False,
    help='Enqueue updated files for processing by the worker')
def poll(enqueue):
    files = tasks.poll(enqueue=enqueue)
    if not enqueue:
        dump_records(files)


@cli.command(help='Clean expired forecast and observations from database')
def clean():
    tasks.clean()


@cli.command(help='Start brightsky worker')
def work():
    from brightsky.worker import huey
    huey.flush()
    config = ConsumerConfig(worker_type='thread', workers=2*cpu_count()+1)
    config.validate()
    consumer = huey.create_consumer(**config.values)
    consumer.run()


@cli.command(help='Start brightsky API webserver')
@click.option('--bind', default='127.0.0.1:5000', help='Bind address')
@click.option(
    '--reload/--no-reload', default=False,
    help='Reload server on source code changes')
def serve(bind, reload):
    from brightsky.web import StandaloneApplication
    StandaloneApplication(
        'brightsky.web:app',
        bind=bind,
        workers=1 if reload else 2*cpu_count()+1,
        reload=reload
    ).run()


@cli.command('query', help='Retrieve weather records')
@click.argument('date', required=False, callback=parse_date_arg)
@click.argument('lat', required=False, type=float)
@click.argument('lon', required=False, type=float)
@click.argument('last-date', required=False, callback=parse_date_arg)
@click.option(
    '--dwd-station-id', help='Query by DWD station ID instead of lat/lon')
@click.option(
    '--wmo-station-id', help='Query by DWD station ID instead of lat/lon')
@click.option(
    '--source-id', type=int, help='Query by source ID instead of lat/lon')
@click.option(
    '--max-dist', type=int, default=50000,
    help='Maximum distance to observation location, in meters')
def query_weather(
        date, lat, lon, last_date, dwd_station_id, wmo_station_id, source_id,
        max_dist):
    if not date:
        date = datetime.datetime.now(tzutc()).replace(
            hour=0, minute=0, second=0, microsecond=0)
    result = query.weather(
        date, last_date=last_date, lat=lat, lon=lon,
        dwd_station_id=dwd_station_id, wmo_station_id=wmo_station_id,
        source_id=source_id, max_dist=max_dist)
    print(json.dumps(result, default=str))


@cli.command('sources', help='Retrieve observation sources')
@click.argument('lat', required=False, type=float)
@click.argument('lon', required=False, type=float)
@click.option(
    '--dwd-station-id', help='Query by DWD station ID instead of lat/lon')
@click.option(
    '--wmo-station-id', help='Query by DWD station ID instead of lat/lon')
@click.option(
    '--source-id', type=int, help='Query by source ID instead of lat/lon')
@click.option(
    '--max-dist', type=int, default=50000,
    help='Maximum distance to observation location, in meters')
def query_sources(
        lat, lon, dwd_station_id, wmo_station_id, source_id, max_dist):
    result = query.sources(
        lat=lat, lon=lon, dwd_station_id=dwd_station_id,
        wmo_station_id=wmo_station_id, source_id=source_id, max_dist=max_dist)
    print(json.dumps(result, default=str))
